@@ -1,6 +1,6 @@
 package Gtk2::Ex::HyperTextBuffer;
 
-our $VERSION = 0.08;
+our $VERSION = '0.15';
 
 use strict;
 use Gtk2;
@@ -29,11 +29,13 @@ instead of plain text. It also keeps an undo-/redo-stack.
 
 # It seems to me that 10 * PANGO_SCALE is the normal font size
 # TODO make tag styles configable
+# FIXME replace "weight" with "scale" ?
 our %TAGS = (
 	bold      => [weight => PANGO_WEIGHT_BOLD],
 	italic    => [style => 'italic'],
 	underline => [underline => 'single'],
-	verbatim  => [family => 'monospace', wrap_mode => 'none'],
+	strike    => [strikethrough => 'true', foreground => 'grey'],
+	verbatim  => [family => 'monospace'], # wrap_mode => 'none'], # Bug withwrap_mode !?
 	Verbatim  => [family => 'monospace', wrap_mode => 'none', indent => 40],
 	link      => [foreground => 'blue', underline => 'single'],
 	head1     => [weight => PANGO_WEIGHT_BOLD, size => 10*1.2**3 * PANGO_SCALE ],
@@ -69,6 +71,13 @@ sub INIT_INSTANCE {
 	my $self = shift;
 	$self->{edit_mode_tags} = [];
 	$self->{image_marks} = [];
+#	$self->signal_connect(mark_set => sub {
+#			my ($self, $iter, $mark) = @_;
+#			return unless $mark->get_name eq 'insert';
+#			my @names = map {$_->get_property('name')} $self->_get_open_tags($iter);
+#			warn "Tags: @names\n";
+#			#$self->set_edit_mode_from_cursor;
+#		} );
 }
 
 =item C<clear()>
@@ -117,12 +126,11 @@ For example:
 		], "\n",
 	]
 
-TODO: add a second method to load Pod::Simple parse trees directly
-	
 =cut
 
 sub set_parse_tree {
 	my ($self, $tree) = @_;
+	#use Data::Dumper; print "HyperTextBuffer got tree: ", Dumper $tree;
 	$self->clear;
 	$self->insert_blocks_at_cursor($tree);
 	$self->set_modified(0);
@@ -176,7 +184,7 @@ sub insert_blocks_at_cursor {
 			}
 			elsif ($type eq 'link') {
 				my $bit = (@$_ == 3 and $$_[2] eq $$meta{to});
-				$tag = $self->create_link_tag([$bit, $$meta{to}]);
+				$tag = $self->create_link_tag($bit ? undef : $$meta{to});
 			}
 			else { $tag = $self->get_tag_table->lookup($type) }
 			
@@ -187,11 +195,19 @@ sub insert_blocks_at_cursor {
 				pop @{$self->{edit_mode_tags}};
 				$_Verbatim-- if $type eq 'Verbatim'; # FIXME
 			}
-			elsif (grep {$_ eq $type} qw/Document Para/) {
-				$self->{base} = $$meta{base} if $$meta{base};
+			elsif ($type eq 'Document' or $type eq 'Para') {
+				if ($$meta{base}) {
+					$self->{base} = $$meta{base};
+					$self->{base} =~ s#^file://##;
+					#warn "base set to: $self->{base}\n";
+				}
 				$self->insert_blocks_at_cursor(@$_[2 .. $#$_]); # recurs
 			}
 			else { die "Unknown tag in parse tree: $type" }
+
+			my $lines = ($type =~ /^head/) ? 1 : 0;
+			$lines += $$meta{empty_lines} if $$meta{empty_lines}; 
+			$self->insert_at_cursor("\n"x$lines) if $lines;
 		}
 	}
 
@@ -209,14 +225,27 @@ sub insert_image_from_file {
 	my $file = $self->{base} ? File::Spec->rel2abs($src, $self->{base}) : $src;
 	#print "file: $$meta{src}, base: $self->{base} => $file\n";
 
-	my $image = Gtk2::Gdk::Pixbuf->new_from_file($file) if -f $file and -r _;
-	$image = Gtk2::Image->new->render_icon('gtk-missing-image', 'button' )
-		unless $image;
-	die "could not insert image: $file\n" unless $image;
+	my $pixbuf = Gtk2::Gdk::Pixbuf->new_from_file($file) if -f $file and -r _;
 
-	$image->{image_src} = $src;
+	$pixbuf ||= Gtk2::Image->new->render_icon('gtk-missing-image', 'dialog')->copy;
+	die "could not insert image: $file\n" unless $pixbuf;
+	
+	if (defined $pixbuf->{image_src}) {
+		# copy does not always copy, or so it seems :(
+		$pixbuf = $pixbuf->copy;
+		warn "WARNING: pixbuf bug !??";
+	}
+	die "ERROR: Pixbuf BUG !" if defined $pixbuf->{image_src};
+	#warn "PixBuf: $pixbuf ($pixbuf->{image_src})\n";
+	
+	$pixbuf->{image_src} = $src;
+	$pixbuf->{image_file} = $file;
 	my $iter = $self->get_iter_at_mark( $self->get_insert );
-	$self->insert_pixbuf($iter, $image);
+
+	#my $mark = $self->create_mark(undef, $iter, 0);
+	#$mark->set_visible(1);
+	
+	$self->insert_pixbuf($iter, $pixbuf);
 
 	return 1;
 }
@@ -230,15 +259,21 @@ Returns a parse tree based on the buffer contents.
 sub get_parse_tree {
 	# TODO Nesting order of tags starting on the same
 	# position is arbitrary - add logic to sort this out
+	# TODO: restructure the tree to order it in paragraphs
 	my ($self, $start, $end) = @_;
 	($start, $end) = $self->get_bounds unless defined $start;
-	my $tree = ['Document', {},];
-	my @stack = ($tree);
+
+	# First we build a parse tree based on the tags we find
+	my $root = [];
+	my @stack = ($root);
 	my $iter = $start->copy;
 	for ($iter->get_tags) {
+		my $link_data = Gtk2::Ex::HyperTextView->get_link_at_iter($iter, $_)
+			if $_->{is_link}; # FIXME quick hack here
 		my $tag = $_->{is_link}
-			? [ 'link', {to => $_->{link_data}[1]} ]
-			: [ $_->get_property('name'), {}       ] ;
+			? [ 'link', {to => $link_data}   ]
+			: [ $_->get_property('name'), {} ] ;
+		next if $$tag[0] =~ /gtkspell/;
 		push @{$stack[-1]}, $tag;
 		push @stack, $tag;
 	}
@@ -267,12 +302,17 @@ sub get_parse_tree {
 		for ($iter->get_toggled_tags(0)) { # closed tags
 			# warning: no consistency check is being done here
 			#          because order of closing tags is unsorted
+			next if !$_->{is_link}
+				and $_->get_property('name') =~ /gtkspell/;
 			pop @stack;
 		}
 		for ($iter->get_toggled_tags(1)) { # opened tags
+			my $link_data = Gtk2::Ex::HyperTextView->get_link_at_iter($iter, $_)
+				if $_->{is_link}; # FIXME quick hack here
 			my $tag = $_->{is_link}
-				? [ 'link', {to => $_->{link_data}[1]} ]
-				: [ $_->get_property('name'), {}       ] ;
+				? [ 'link', {to => $link_data}   ]
+				: [ $_->get_property('name'), {} ] ;
+			next if $$tag[0] =~ /gtkspell/;
 			push @{$stack[-1]}, $tag;
 			push @stack, $tag;
 		}
@@ -282,19 +322,55 @@ sub get_parse_tree {
 		goto EXTRACT_TAGS;
 	}
 
-	#use Data::Dumper; print STDERR Dumper $tree;
-	return $tree;
+	# Check lines around headings - FIXME very ugly implementation
+	# Also add sanity check to merge blocks of 'verbatim' to 'Verbatim'
+	for (0 .. $#$root) {
+		next unless ref $$root[$_] and $$root[$_][0] =~ /^head/;
+		$$root[$_][-1] =~ s/(\n*)$//;
+		my $lines = length $1;
+
+		# check empty lines at start next para
+		if (ref $$root[$_+1]) { # unwrap tree
+			my $next = $$root[$_+1];
+			next if $$next[0] =~ /^head/;
+			while (ref $$next[2]) { $next = $$next[2] }
+			$$next[2] =~ s/^(\n*)//;
+			$lines += length $1;
+			splice @$next, 2, 1 unless length $$next[2];
+			$$root[$_+1] = undef unless @{$$root[$_+1]} > 2;
+		}
+		elsif (defined $$root[$_+1]) {
+			$$root[$_+1] =~ s/^(\n*)//;
+			$lines += length $1;
+			$$root[$_+1] = undef unless length $$root[$_+1];
+		}
+		
+		# Set end of line on previous para
+		if ($_ > 0) {
+			if (ref $$root[$_-1]) { # unwrap tree
+				my $prev = $$root[$_-1];
+				next if $$prev[0]  =~ /^head/;
+				while (ref $$prev[-1]) { $prev = $$prev[-1] }
+				$$prev[-1] =~ s/\n?$/\n/;
+			}
+			elsif (defined $$root[$_-1]) {
+				$$root[$_-1] =~ s/\n?$/\n/;
+			}
+		}
+		
+		$$root[$_][1]{empty_lines} = $lines - 1 if $lines > 1;
+	}
+	@$root = grep defined($_), @$root;
+		
+	unshift @$root, 'Document', {};
+	#use Data::Dumper; print STDERR Dumper $root;
+	return $root;
 }
 
 =item C<create_link_tag(LINK_DATA)>
 
 Returns an anonymous L<Gtk2::TextIter> that holds LINK_DATA.
 
-LINK_DATA should be an array reference. The first item in this array
-should be a boolean descibing whether the link matches the text it is applied
-to. If this boolean is set to tru the second element in the array will be update
-to hold the text if the buffer is edited.
-
 Every link tag should only be applied to one text region.
 
 =cut
@@ -309,22 +385,115 @@ sub create_link_tag {
 	return $tag;
 }
 
-=item C<select_word()>
+=item C<auto_selection(TAG, ITER)>
 
-TODO
+Returns start and end iterators for the current selection.
+If there is no selection in the buffer TAG and ITER are used to
+automaticly select a range if possible.
+ITER is optional and defaults to the cursor.
+
+=cut
+
+sub auto_selection {
+	my ($self, $tag, $iter) = @_;
+	my ($start, $end) = $self->get_selection_bounds;
+	#warn "selection: $start, $end tag: $tag\n";
+	if (!$end or $start == $end) { # autoselect word if defined tag
+		if ($tag eq 'link') {
+			($start, $end) = $self->select_link($iter);
+			($start, $end) = $self->select_word($iter)
+				if !$end or $start == $end;
+		}
+		elsif ($tag =~ /^head/) {
+			($start, $end) = $self->select_line($iter);
+		}
+		elsif ($tag and $tag !~ /^[A-Z]/) {
+			($start, $end) = $self->select_word($iter);
+		}
+		#warn "   => $start, $end\n";
+		return undef if !$end or $start == $end;
+	}
+	return ($start, $end);
+}
+
+=item C<select_word(ITER)>
+
+Selects the current word if ITER is inside a word.
+Cursor is used when ITER is undefined.
+Returns Gtk2::TextIter for start and end or undef.
 
 =cut
 
 sub select_word {
+	my ($self, $iter) = @_;
+	$iter ||= $self->get_iter_at_mark( $self->get_insert );
+	#warn "select-word: ", $iter->inside_word, "\n";
+	return unless $iter->inside_word
+		or $iter->starts_word or $iter->ends_word;
+	my $begin = $iter->copy;
+	$begin->backward_word_start unless $begin->starts_word;
+	$iter->forward_word_end unless $iter->ends_word;
+	$self->select_range($begin, $iter);
+	return ($begin, $iter);
 }
 
-=item C<select_line>
+=item C<select_line(ITER)>
 
-TODO
+Selects the line that contains ITER.
+Cursor is used when ITER is undefined.
+Returns Gtk2::TextIter for start and end or undef.
 
 =cut
 
 sub select_line {
+	my ($self, $iter) = @_;
+	$iter ||= $self->get_iter_at_mark( $self->get_insert );
+	my $begin = $iter->copy;
+	$begin = $self->get_iter_at_line( $begin->get_line )
+		unless $begin->starts_line;
+	while (! $iter->ends_line) { $iter->forward_char }
+	$self->select_range($begin, $iter);
+	return ($begin, $iter);
+}
+
+=item C<select_tag(ITER, TAG)>
+
+Selects TAG if ITER has TAG.
+Cursor is used when ITER is undefined.
+Tag can be either a Gtk2::TextTag or a tag name.
+Returns Gtk2::TextIter for start and end or undef.
+
+=cut
+
+sub select_tag {
+	my ($self, $iter, $tag) = @_;
+	#warn "select_tag: $tag\n";
+	$tag = $self->get_tag_table->lookup($tag) unless ref $tag;
+	$iter ||= $self->get_iter_at_mark( $self->get_insert );
+	return unless $iter->has_tag($tag) or $iter->toggles_tag($tag);
+	my $begin = $iter->copy;
+	$begin->backward_to_tag_toggle($tag) unless $begin->begins_tag($tag);
+	$iter->forward_to_tag_toggle($tag) unless $begin->ends_tag($tag);
+	$self->select_range($begin, $iter);
+	return ($begin, $iter);
+}
+
+=item C<select_link(ITER)>
+
+Selects link if ITER is on a link.
+Cursor is used when ITER is undefined.
+Returns Gtk2::TextIter for start and end or undef.
+
+=cut
+
+sub select_link {
+	#warn "select_link\n";
+	my ($self, $iter) = @_;
+	$iter ||= $self->get_iter_at_mark( $self->get_insert );
+	for ($iter->get_tags) {
+		next unless $_->{is_link};
+		return $self->select_tag($iter, $_);
+	}
 }
 
 =item C<apply_tag_to_selection>
@@ -338,13 +507,28 @@ sub apply_tag_to_selection {
 	
 }
 
-=item C<replace_selection>
+=item C<replace_selection(TEXT)>
 
-TODO
+Replace any selected text with TEXT.
+Returns iterators for the start and end of the newly inserted selection.
 
 =cut
 
 sub replace_selection {
+	my ($self, $text) = @_;
+	my ($start, $end) = $self->get_selection_bounds;
+	return if !$end or $start == $end;
+
+	my $_start = $start->get_offset;
+	my $_end   = $start->get_offset + length $text;
+
+	$self->delete($start, $end);
+	$self->insert($start, $text);
+
+	($start, $end) = map $self->get_iter_at_offset($_), ($_start, $_end);
+	$self->select_range($start, $end);
+
+	return $start, $end;
 }
 
 =item C<set_edit_mode_tags>
@@ -366,11 +550,23 @@ TODO
 
 =cut
 
+sub _get_open_tags {
+	# iter->get_tags has "left" gravity
+	# this function has "right" gravity
+	my ($self, $iter) = @_;
+	my @tags = $iter->get_tags;
+	for my $t ($iter->get_toggled_tags(1)) { # tags that are toggles on here
+		@tags = grep {$_ ne $t} @tags;
+	}
+	push @tags, $iter->get_toggled_tags(0);
+	return @tags;
+}
+
 sub set_edit_mode_from_cursor {
 	my $self = shift;
 	my $iter = $self->get_iter_at_mark( $self->get_insert );
 	$self->{edit_mode_tags} = [$iter->get_tags];
-	#print join ' ', "Mode: ", map(
+	#print join ' ', "Mode:", map(
 	#	{$_->get_property('name')} @{$self->{edit_mode_tags}}
 	#), "\n";
 }
@@ -525,7 +721,7 @@ sub paste_clipboard {
 	my ($start, $end) = $self->get_selection_bounds;
 	if (defined $start) {
 		my $iter = $insert || $self->get_iter_at_mark( $self->get_insert );
-		if ($iter->in_range($start, $end)) {
+		if ($iter->in_range($start, $end) or $iter->equal($end)) {
 			$insert = $iter;
 			$replace = 1;
 		}
@@ -581,7 +777,7 @@ sub _clipboard_paste_data {
 	# check selection
 	$self->delete_selection(@{$paste_data}{'interactive','editable'})
 		if $paste_data->{replace};
-
+	
 	# check gtk_paste_point_override
 	my $iter;
 	my $mark = $self->get_mark('gtk_paste_point_override');
