@@ -1,6 +1,6 @@
 package Gtk2::Ex::HyperTextView;
 
-our $VERSION = '0.08';
+our $VERSION = '0.15';
 
 use strict;
 use Gtk2;
@@ -44,8 +44,22 @@ but also with other TextBuffer classes.
 
 =over 4
 
+=item C<new()>
+
+=item C<new_with_buffer(BUFFER)>
+
 =cut
 
+# sub new() inherited from Glib::Object
+
+sub new_with_buffer {
+	# needs to be overoaded, TextView->new_with returns TextView object
+	my ($class, $buffer) = @_;
+	my $self = $class->new();
+	$self->set_buffer($buffer);
+	return $self;
+}
+
 sub INIT_INSTANCE {
 	my $self = shift;
 	
@@ -143,9 +157,8 @@ mouse cursor, not at the text cursor.
 
 sub get_link_at_pointer {
 	my $self = shift;
-	my ($x, $y) = $self->get_pointer;
-	   ($x, $y) = $self->window_to_buffer_coords('widget', $x, $y);
-	my $iter = $self->get_iter_at_location($x, $y);
+	my $iter = $self->get_iter_at_pointer;
+	return unless defined $iter;
 	return $self->get_link_at_iter($iter);
 }
 
@@ -155,11 +168,18 @@ Returns link data or undef.
 
 =cut
 
-sub get_link_at_iter {
-	my ($self, $iter) = @_;
-	my ($link_tag) = grep {$_->{is_link}} $iter->get_tags;
-	return undef unless $link_tag;
-	return $link_tag->{link_data};
+sub get_link_at_iter { # $tag is a private argument
+	my ($self, $iter, $tag) = @_;
+	($tag) = grep {$_->{is_link}} $iter->get_tags unless $tag;
+	return undef unless $tag;
+	my $link_data = $tag->{link_data};
+	unless (defined $link_data) {
+		my ($begin, $end) = ($iter->copy, $iter->copy);
+		$begin->backward_to_tag_toggle($tag) unless $begin->begins_tag($tag);
+		$end->forward_to_tag_toggle($tag) unless $end->ends_tag($tag);
+		$link_data = $begin->get_text($end);
+	}
+	return $link_data;
 }
 
 =item click_if_link_at_cursor( )
@@ -188,15 +208,28 @@ Returns undef or link data.
 sub click_if_link_at_iter {
 	my ($self, $iter) = @_;
 
-	my ($link_tag) = grep {$_->{is_link}} $iter->get_tags;
-	return undef unless $link_tag;
-	my $link_data = $link_tag->{link_data};
+	my $link_data = $self->get_link_at_iter($iter);
+	return unless defined $link_data;
 
 	$self->signal_emit('link_clicked', $link_data);
 
 	return $link_data;
 }
 
+=item C<get_iter_at_pointer()>
+
+Returns the text iter for the current pointer (mouse cursor) position.
+
+=cut
+
+sub get_iter_at_pointer {
+	my $self = shift;
+	my ($x, $y) = $self->get_pointer;
+	   ($x, $y) = $self->window_to_buffer_coords('widget', $x, $y);
+	return unless defined $x;
+	my $iter = $self->get_iter_at_location($x, $y);
+}
+
 =item set_cursor_if_appropriate(X, Y)
 
 Looks at all tags covering the position (X, Y) in the text view, 
@@ -211,15 +244,14 @@ Returns the link data if the cursor is above a link.
 
 =cut
 
-sub set_cursor_if_appropriate { # FIXME add default x, y
+sub set_cursor_if_appropriate {
 	my ($self, $x, $y) = @_;
 	
-	unless (defined $x) {
-		($x, $y) = $self->get_pointer;
-		($x, $y) = $self->window_to_buffer_coords('widget', $x, $y);
-	}
-	
-	my $iter = $self->get_iter_at_location($x, $y);
+	my $iter = defined($x) 
+		? $self->get_iter_at_location($x, $y)
+		: $self->get_iter_at_pointer()        ;
+	return unless defined $iter;
+
 	my ($link_tag) = grep {$_->{is_link}} $iter->get_tags;
 
 	my $hovering = $link_tag ? 1 : 0;
@@ -231,7 +263,7 @@ sub set_cursor_if_appropriate { # FIXME add default x, y
 		$self->get_window('text')->set_cursor($cursor);
 	}
 	
-	return $hovering ? $link_tag->{link_data} : undef;
+	return $hovering ? $self->get_link_at_iter($iter, $link_tag) : undef;
 }
 
 # ###### #
@@ -304,15 +336,18 @@ sub on_button_release_event {
 	my ($start, $end) = $self->get_buffer->get_selection_bounds;
 	return 0 if $start && $end and $start->get_offset != $end->get_offset;
 	
-	my ($x, $y) = $event->get_coords;
-	   ($x, $y) = $self->window_to_buffer_coords('widget', $x, $y);
-	my $iter = $self->get_iter_at_location($x, $y);
+	my $iter = $self->get_iter_at_pointer;
+	return 0 unless defined $iter;
 
 	$self->click_if_link_at_iter($iter);
 
 	return 0;
 }
 
+# The following three methods are needed because TextBuffer doesn't have
+# signals for these things and we want to make sure the overloaded methods
+# in HyperTextBuffer get called.
+
 sub on_copy_clipboard {
 	my $view = shift;
 	my $buffer = $view->get_buffer;
